;; Generated package description from pulsar.el  -*- no-byte-compile: t -*-
(define-package "pulsar" "0.5.0.0.20230809.50958" "Pulse highlight on demand or after select functions" '((emacs "27.1")) :commit "aa0969301880ec53e0657b8ca19ba770011eb5a9" :authors '(("Protesilaos Stavrou" . "info@protesilaos.com")) :maintainer '("Pulsar Development" . "~protesilaos/pulsar@lists.sr.ht") :keywords '("convenience" "pulse" "highlight") :url "https://git.sr.ht/~protesilaos/pulsar")
